#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <math.h>
#include <ctype.h>
#include <string.h>

#define windowX 0
#define windowY 0
#define border  1
#define dscreen DefaultScreen(display)

Display *display;
Window window;
XColor color[16];
unsigned width,height;

long *oldh;

int colmap[16][3]={0,0,0,
                   15,0,0,
                   15,11,0,
                   15,14,4,
                   14,13,9,
                   13,13,12,
                   12,12,13,
                   11,11,12,
                   10,10,11,
                   9,9,10,
                   8,8,9,
                   7,7,8,
                   6,6,7,
                   5,5,6,
                   4,4,5,
                   3,3,4};

double MSetPot(cx,cy,maxiter)
double cx,cy;
long maxiter;
{
	long iter=0;
	register double x,y,x2,y2,temp,pot=0.0;

	for (x=cx,y=cy,x2=x*x,y2=y*y;(iter<maxiter)&&(x2+y2<10000.0);iter++)
	{
		temp=x2-y2+cx;
		y=2*x*y+cy;
		x=temp;
		x2=x*x;
		y2=y*y;
	}

	if (iter<maxiter) pot=0.5*log(x2+y2)/pow(2.0,(double)iter);
	return(pot);
}


MSetCPM3D(xmin,xmax,ymin,ymax,maxiter,angle,scale,lift)
double xmin,xmax,ymin,ymax,scale,lift;
long maxiter,angle;
{
	register long ix,iy,shx,shh,shdep,col,h,maxh,cut,allsteps=0;
	double delta,cx,cy,pot,cutf;

	maxh=height;
	if (angle<0)
	{
		allsteps=1;
		angle = -angle;
	}
	cut=maxh/2;
	if (ymax-ymin!=xmax-xmin) ymax=ymin+xmax-xmin;
        XSetForeground(display,DefaultGC(display,dscreen),color[1].pixel);
	for (iy=0;iy<angle*cut;)
	{
		cy=ymin+iy*(ymax-ymin)/width;
		for (ix=0;ix<width;ix++)
		{
			cx=xmin+ix*(xmax-xmin)/width;
			pot=(double)MSetPot(cx,cy,maxiter);
			h=maxh*(1-pow(pow(2.0,scale)*pot,lift));
			if (iy>0)
			{
				col=(oldh[ix]+h)/2-oldh[ix-1]+8;
				if (col<4) col=8;
				if (col<6) col=7;
				if ((oldh[ix]>=cut)&&(h>=cut))
				{
					col=8;
					if (oldh[ix]>cut+(maxh-cut)/8) col=7;
					if (oldh[ix]>cut+(maxh-cut)/6) col=6;
					if (oldh[ix]>cut+(maxh-cut)/4) col=5;
					if (oldh[ix]>cut+(maxh-cut)/2) col=4;
					if (oldh[ix]>cut+3*(maxh-cut)/4) col=3;
					if (oldh[ix]>cut+5*(maxh-cut)/6) col=2;
					if (oldh[ix]>cut+7*(maxh-cut)/8) col=1;
					oldh[ix]=cut;
				}
				if (oldh[ix]>=cut) oldh[ix]=cut;
				if (oldh[ix]<cut)
				{
					shh=oldh[ix]-ix;
					for (shdep=0,shx=ix;(shx<width)&&(shx+shh<cut)&&(shdep<2);shx++)
						if (shx+shh<oldh[shx]) shdep++;
					col+=shdep;
				}
				if (col>15) col=15;
				if (pot==0.0) col=0;
				if (ix>0)
				{
            			XSetForeground(display,DefaultGC(display,dscreen),color[col].pixel);
                                XDrawLine(display,window,DefaultGC(display,dscreen),ix,cut+(iy-1)/angle,ix,cut+(iy-1)/angle-oldh[ix]);
                                }
			}
			oldh[ix]=h;
		}
		if (allsteps==0) iy+=angle;
		else iy++;
	}
		for (ix=1;ix<width;ix++)
		{
            	XSetForeground(display,DefaultGC(display,dscreen),color[9].pixel);
                if(oldh[ix]>cut) oldh[ix]=cut;
                XDrawLine(display,window,DefaultGC(display,dscreen),ix,height/2+(iy-1)/angle,ix,height/2+(iy-1)/angle-oldh[ix]+1);
		}
}


main(int argc,char **argv)
{
int i,j,flag,gotcolors,end=0,maxit,angle,own=0;
double xmin,xmax,ymin,ymax,scale,lift;

unsigned long attribute_mask;
XSetWindowAttributes window_attributes;

Colormap cmap;
unsigned long pmr,pir;

if (argc==1) DisplayUsage();
if ((argc==2)&&(strcmp(argv[1],"?")==0)) DisplayUsage();
        i=1;
        switch (argc)
        {
                case 12:
                        own = 1;
                        i++;
                case 11:
                        width=atoi(argv[i++]);
                        height=atoi(argv[i++]);
                        xmin=atof(argv[i++]);
                        xmax=atof(argv[i++]);
                        ymin=atof(argv[i++]);
                        ymax=atof(argv[i++]);
                        maxit=atoi(argv[i++]);
                        angle=atoi(argv[i++]);
                        scale=atof(argv[i++]);
                        lift=atof(argv[i++]);
                        break;
                default: DisplayUsage();
        }

if(!(display = XOpenDisplay(NULL)))
  {
  puts("Can't open Display !");
  exit(0);
  }

if(own)
  cmap = XCreateColormap(display,RootWindow(display,dscreen),DefaultVisual(display,dscreen),0);
else
  cmap = DefaultColormap(display,dscreen);

window_attributes.border_pixel = BlackPixel(display,dscreen);
window_attributes.background_pixel = WhitePixel(display,dscreen);
window_attributes.event_mask = KeyPressMask|ExposureMask;
window_attributes.backing_store = Always;
window_attributes.colormap = cmap;

attribute_mask = CWBorderPixel|CWBackPixel|CWEventMask|CWBackingStore|CWColormap;

window = XCreateWindow( display,
                        RootWindow(display,dscreen),
                        windowX,
                        windowY,
                        width,
                        height,
                        border,
                        DefaultDepth(display,dscreen),
                        CopyFromParent,
                        CopyFromParent,
                        attribute_mask,
                        &window_attributes);
  	
XChangeProperty(display,window,
                XA_WM_NAME,
                XA_STRING,
                8,
                PropModeReplace,
                "CPM",
                3);

gotcolors = 0;

for(i=0;i<16;i++)
  {
  color[i].red   = 4369*colmap[i][0];
  color[i].green = 4369*colmap[i][1];
  color[i].blue  = 4369*colmap[i][2];
  color[i].flags = DoRed|DoGreen|DoBlue;

  XAllocColor(display,cmap,&color[i]);

  flag = 0;
  for(j=0;j<i;j++) if(color[j].pixel==color[i].pixel) flag = 1;
  if(flag)
    flag=0;
  else
    gotcolors++;
  }

XMapWindow(display,window);                        

printf("Could allocate %d colors\n",gotcolors);
printf("Doing: Width %d Height %d Maxit %ld Angle %ld Scale %2.0f Lift %1.5f\n",width,height,maxit,angle,scale,lift);
oldh=malloc(width*sizeof(long));
MSetCPM3D(xmin,xmax,ymin,ymax,maxit,angle,scale,lift);

XFlush(display);
   	
while(1)
  {
  XEvent event;
  XNextEvent(display,&event);
  switch(event.type)
    {
    case KeyPress : exit(0);
    }
  }
}

DisplayUsage(void)
{
        puts("CPM - 3D Mandelbrot program using Continuous Potiential Method");
        puts("      1989 by Lars R. Clausen (2:230/22.34)");
        puts("      1992 by Stefan Blomen (X-Windows - Implementation)");
        puts("      Algorithm from 'The Science of Fractal Images'.");
        puts("Usage: CPM [-own] width height xmin xmax ymin ymax maxit angle scale lift");
        exit(0);
}
